@@ -10,8 +10,8 @@ use crate::config::{ClientAddrConfig, TorClientConfig};
 use tor_circmgr::{IsolationToken, StreamIsolationBuilder, TargetPort};
 use tor_dirmgr::DirEvent;
 use tor_persist::{FsStateMgr, StateMgr};
-use tor_proto::circuit::{ClientCirc, IpVersionPreference};
-use tor_proto::stream::DataStream;
+use tor_proto::circuit::ClientCirc;
+use tor_proto::stream::{DataStream, IpVersionPreference, StreamParameters};
 use tor_rtcompat::{Runtime, SleepProviderExt};
 
 use futures::stream::StreamExt;
@@ -54,6 +54,8 @@ pub struct ConnectPrefs {
     ip_ver_pref: IpVersionPreference,
     /// Id of the isolation group the connection should be part of
     isolation_group: Option<IsolationToken>,
+    /// Whether to return the stream optimistically.
+    optimistic_stream: bool,
 }
 
 impl ConnectPrefs {
@@ -98,10 +100,24 @@ impl ConnectPrefs {
         self
     }
 
-    /// Get the begin_flags fields that we should use for the BEGIN
-    /// cell for this stream.
-    fn begin_flags(&self) -> IpVersionPreference {
-        self.ip_ver_pref
+    /// Indicate that the stream should be opened "optimistically".
+    ///
+    /// By default, streams are not "optimistic". When you call
+    /// [`TorClient::connect()`], it won't give you a stream until the
+    /// exit node has confirmed that it has successfully opened a
+    /// connection to your target address.  It's safer to wait in this
+    /// way, but it is slower: it takes an entire round trip to get
+    /// your confirmation.
+    ///
+    /// If a stream _is_ configured to be "optimistic", on the other
+    /// hand, then `TorClient::connect()` will return the stream
+    /// immediately, without waiting for an answer from the exit.  You
+    /// can start sending data on the stream right away, though of
+    /// course this data will be lost if the connection is not
+    /// actually successful.
+    pub fn optimistic(&mut self) -> &mut Self {
+        self.optimistic_stream = true;
+        self
     }
 
     /// Return a TargetPort to describe what kind of exit policy our
@@ -113,6 +129,15 @@ impl ConnectPrefs {
         }
     }
 
+    /// Return a new StreamParameters based on this configuration.
+    fn stream_parameters(&self) -> StreamParameters {
+        let mut params = StreamParameters::default();
+        params
+            .ip_version(self.ip_ver_pref)
+            .optimistic(self.optimistic_stream);
+        params
+    }
+
     /// Indicate which other connections might use the same circuit
     /// as this one.
     pub fn set_isolation_group(&mut self, isolation_group: IsolationToken) -> &mut Self {
@@ -134,16 +159,10 @@ impl<R: Runtime> TorClient<R> {
     ///
     /// Return a client once there is enough directory material to
     /// connect safely over the Tor network.
-    pub async fn bootstrap(
-        runtime: R,
-        TorClientConfig {
-            state_cfg,
-            dir_cfg,
-            circ_cfg,
-            addr_cfg,
-        }: TorClientConfig,
-    ) -> Result<TorClient<R>> {
-        let statemgr = FsStateMgr::from_path(state_cfg)?;
+    pub async fn bootstrap(runtime: R, config: TorClientConfig) -> Result<TorClient<R>> {
+        let circ_cfg = config.get_circmgr_config()?;
+        let dir_cfg = config.get_dirmgr_config()?;
+        let statemgr = FsStateMgr::from_path(config.storage.expand_state_dir()?)?;
         if statemgr.try_lock()?.held() {
             debug!("It appears we have the lock on our state files.");
         } else {
@@ -151,6 +170,7 @@ impl<R: Runtime> TorClient<R> {
                 "Another process has the lock on our state files. We'll proceed in read-only mode."
             );
         }
+        let addr_cfg = config.address_filter.clone();
         let chanmgr = Arc::new(tor_chanmgr::ChanMgr::new(runtime.clone()));
         let circmgr =
             tor_circmgr::CircMgr::new(circ_cfg, statemgr.clone(), &runtime, Arc::clone(&chanmgr))?;
@@ -197,7 +217,7 @@ impl<R: Runtime> TorClient<R> {
     /// Return a new isolated `TorClient` instance.
     ///
     /// The two `TorClient`s will share some internal state, but their
-    /// streams will haver share circuits with one another.
+    /// streams will never share circuits with one another.
     ///
     /// Use this function when you want separate parts of your program to
     /// each have a TorClient handle, but where you don't want their
@@ -234,7 +254,8 @@ impl<R: Runtime> TorClient<R> {
         // TODO: make this configurable.
         let stream_timeout = Duration::new(10, 0);
 
-        let stream_future = circ.begin_stream(&addr, port, Some(flags.begin_flags()));
+        let stream_future = circ.begin_stream(&addr, port, Some(flags.stream_parameters()));
+        // This timeout is needless but harmless for optimistic streams.
         let stream = self
             .runtime
             .timeout(stream_timeout, stream_future)
@@ -405,7 +426,7 @@ async fn update_persistent_state<R: Runtime>(
                 Ok(NewlyAcquired) => {
                     info!("We now own the lock on our state files.");
                     if let Err(e) = circmgr.upgrade_to_owned_persistent_state() {
-                        error!("Unable to upgrade to owneed state files: {}", e);
+                        error!("Unable to upgrade to owned state files: {}", e);
                         break;
                     }
                 }
@@ -460,7 +481,7 @@ async fn continually_launch_timeout_testing_circuits<R: Runtime>(
         if let (Some(cm), Some(dm)) = (Weak::upgrade(&circmgr), Weak::upgrade(&dirmgr)) {
             let netdir = dm.netdir();
             if let Err(e) = cm.launch_timeout_testing_circuit_if_appropriate(&netdir) {
-                warn!("Problem launching a timeout testing circuit: {}", e)
+                warn!("Problem launching a timeout testing circuit: {}", e);
             }
             delay = netdir
                 .params()
@@ -482,7 +503,7 @@ impl<R: Runtime> Drop for TorClient<R> {
         match self.circmgr.store_persistent_state() {
             Ok(()) => info!("Flushed persistent state at exit."),
             Err(tor_circmgr::Error::State(tor_persist::Error::NoLock)) => {
-                debug!("Lock not held; no state to flush.")
+                debug!("Lock not held; no state to flush.");
             }
             Err(e) => error!("Unable to flush state on client exit: {}", e),
         }
